/*
    Gens/Genecyst save-state hacking tool for the legendary Sega Genesis game Phantasy Star IV
    Copyright (C) 2006-2007, Free Console Organization

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/


#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

typedef unsigned long  u32_t;
typedef unsigned short u16_t;
typedef unsigned char  u8_t;

typedef signed long    s32_t;
typedef signed short   s16_t;
typedef signed char    s8_t;

#define NCHARS   11
#define BASE_ADR 0x11981

#define MESETA_ADR 0x118B1

#define PACKAGE "psiv-hack"
#define VERSION "0.2.0"

typedef struct char_data_s
{
	u8_t level[1];
	u8_t unused0[1];
	u8_t exp[3];
	u8_t cur_hp[2];
	u8_t max_hp[2];
	u8_t cur_tp[2];
	u8_t max_tp[2];
	u8_t unused1[2];
	u8_t str[3];
	u8_t mental[3];
	u8_t agi[3];
	u8_t dex[3];
	u8_t unused3[40];
	u8_t weapons[2];
	u8_t helmet[1];
	u8_t armor[1];
	u8_t unused4[26];
	u8_t tech[10];
	u8_t unused5[21];
} char_data_t;


typedef struct char_input_s
{
	char *agi;
	char *armor;
	char *dex;
	char *exp;
	char *helmet;
	char *hp;
	char *level;
	char *mental;
	char *name;
	char *str;
	char *tp;
	char *weapons;
} char_input_t;

char_data_t char_datas[NCHARS];

typedef struct char_s
{
	int id;
	const char *name;
	char_data_t *data;
} char_t;

char_t chars[NCHARS] = {
	{0, "Chaz", &char_datas[0]},
	{1, "Alys", &char_datas[1]},
	{2, "Hahn", &char_datas[2]},
	{3, "Rune", &char_datas[3]},
	{4, "Gryz", &char_datas[4]},
	{5, "Rika", &char_datas[5]},
	{6, "Demi", &char_datas[6]},
	{7, "Wren", &char_datas[7]},
	{8, "Raja", &char_datas[8]},
	{9, "Kyra", &char_datas[9]},
	{10, "Seth", &char_datas[10]},
};

unsigned int read_num(void *d, int b)
{
	int n=0, i;
	unsigned char *p = (unsigned char*)d + (b-1);


	for(i=0; i<(b<<3); i+=8) n |= (unsigned long)(*p--) << i;
	
	return n;
}

void write_num(unsigned int n, void *d, int b)
{
	int i;
	unsigned char *p = (unsigned char*)d + (b-1);

	for(i=0; i<(b<<3); i+=8) *p-- = (unsigned char)((n >> i)&0xff);
}

void print_char(FILE *fp, char_t *c)
{
	fprintf(fp, "name:\t\t %s\n", c->name);
	fprintf(fp, "level:\t\t %u\n", read_num(c->data->level,1));
	fprintf(fp, "exp:\t\t %u\n", read_num(c->data->exp, 3));
	fprintf(fp, "hp:\t\t %u/%u\n", read_num(c->data->cur_hp, 2), read_num(c->data->max_hp, 2));
	fprintf(fp, "tp:\t\t %u/%u\n", read_num(c->data->cur_tp, 2), read_num(c->data->max_tp, 2));
	fprintf(fp, "str:\t\t %u\n", read_num(c->data->str, 1));
	fprintf(fp, "mental:\t\t %u\n", read_num(c->data->mental, 1));
	fprintf(fp, "agi:\t\t %u\n", read_num(c->data->agi, 1));
	fprintf(fp, "dex:\t\t %u\n", read_num(c->data->dex, 1));
	fprintf(fp, "helmet:\t\t 0x%x\n", read_num(&c->data->helmet,1));
	fprintf(fp, "armor:\t\t 0x%x\n", read_num(&c->data->armor,1));
	fprintf(fp, "weapon1:\t 0x%x\n", read_num(&c->data->weapons[0],1));
	fprintf(fp, "weapon2:\t 0x%x\n", read_num(&c->data->weapons[1],1));
}

char_t *find_char(const char *name)
{
	int i;
	for(i=0; i<NCHARS; i++)
	{
		if(!strcmp(name, chars[i].name)) return &chars[i];
	}
	
	return NULL;
}

int set_char(char_input_t *data, char_t *c)
{
	int n,m;
	char *endptr;
	
	assert(data);
	assert(c);
		
#define _READN(x) ((x) ? strtod(x,0) : -1)

	n = _READN(data->agi);
	if(n>=0) write_num(n, c->data->agi,3);
	
	n = _READN(data->dex);
	if(n>=0) write_num(n, c->data->dex,3);
	
	n = _READN(data->exp);
	if(n>=0) write_num(n, c->data->exp,3);
	
	n = _READN(data->level);
	if(n>=0) write_num(n, c->data->level,1);
	
	n = _READN(data->mental);
	if(n>=0) write_num(n, c->data->mental,3);
	
	n = _READN(data->str);
	if(n>=0) write_num(n, c->data->str,3);

	n = _READN(data->helmet);
	if(n>=0) write_num(n, c->data->helmet,1);
	
	n = _READN(data->armor);
	if(n>=0) write_num(n, c->data->armor,1);

#define _READN_ENDPTR(x) ((x) ? strtod(x,&endptr) : -1)
#define _READN2(x) do { n = _READN_ENDPTR(x); endptr++; m = _READN(endptr); } while(0)

	if(data->hp)
	{
		_READN2(data->hp);
		
		if(n>=0) write_num(n, c->data->cur_hp,2);
		if(m>=0) write_num(n, c->data->max_hp,2);
	}
	
	if(data->tp)
	{
		_READN2(data->tp);
		
		if(n>=0) write_num(n, c->data->cur_tp,2);
		if(m>=0) write_num(n, c->data->max_tp,2);
	}
		
	
	if(data->weapons)
	{
		_READN2(data->weapons);
		
		if(n>=0) write_num(n, &c->data->weapons[0],1);
		if(m>=0) write_num(n, &c->data->weapons[1],1);
	}

	return 0;

#undef _READN
#undef _READN2
}

void license()
{
	puts("You may redistribute copies of this program");
	puts("under the terms of the GNU General Public License.");
	puts("For more information about these matters, see the file named COPYING.");
}

void usage()
{
	fprintf(stdout, "%s %s\n\n", PACKAGE, VERSION);
	
	fprintf(stdout, "usage: %s -n <name> [-a agi -A armor -d dex -e exp -h curhp,maxhp -H helmet -l level -m mental -s str -t curtp,maxtp -W weapon1,weapon2] [-M mesetas] [-o <output>] <savestatefile>\n\n\n", PACKAGE);
	
	license();
}



int main(int argc, char *argv[])
{
	FILE *fp;
	char c;
	char *name=NULL;
	char *outfile=NULL;
	int opt_print_char=0;
	char_t *cp=NULL;
	char *savefile;
	long filesize;
	char_input_t in;
	char *in_mesetas = NULL;
	
	memset(&in, 0, sizeof(in));
	
	if(argc < 2)
	{
		usage();
		return 0;
	}

	while ((c = getopt (argc, argv, "a:A:d:e:h:H:l:m:n:o:ps:t:W:M:")) != -1)
	{
		switch (c)
		{
		case 'n':
			name = optarg;
		break;
		
		case 'a':
			in.agi = optarg;
		break;
		
		case 'A':
			in.armor = optarg;
		break;
		
		case 'd':
			in.dex = optarg;
		break;
		
		case 'e':
			in.exp = optarg;
		break;
		
		case 'h':
			in.hp = optarg;
		break;
		
		case 'H':
			in.helmet = optarg;
		break;
		
		case 'l':
			in.level = optarg;
		break;
		
		case 'm':
			in.mental = optarg;
		break;
		
		case 's':
			in.str = optarg;
		break;
		
		case 't':
			in.tp = optarg;
		break;
		
		case 'W':
			in.weapons = optarg;
		break;
		
		case 'p':
			opt_print_char = 1;
		break;
		
		case 'M':
			in_mesetas = optarg;
		break;
		
		case 'o':
			outfile = optarg;
		break;
		}
	}


	/* last checks for args */
	if(!name)
	{
		usage();
		exit(1);
	}
	
	cp = find_char(name);
	if(!cp)
	{
		fprintf(stderr, "error: no such char: %s\n", name);
		exit(1);
	}

	if(!outfile) outfile = argv[optind];


	/* read char datas */
	fp = fopen(argv[optind], "rb");
	assert(fp);
	fseek(fp, BASE_ADR, SEEK_SET);
	fread(char_datas, sizeof(char_data_t), NCHARS, fp);
	
	fseek(fp, 0, SEEK_END);
	filesize = ftell(fp);
	rewind(fp);


	/* read entire savefile */
	savefile = malloc(filesize);
	if(!savefile)
	{
		fprintf(stderr, "error: cannot allocate memory!\n");
		exit(1);
	}
	fread(savefile, filesize, 1, fp);
	fclose(fp);

	if(opt_print_char) {
		print_char(stdout, cp);
		fprintf(stdout, "mesetas:\t %u\n", read_num(savefile+MESETA_ADR,3));
	}
	
	set_char(&in, cp);
	
	memcpy(savefile+BASE_ADR, char_datas, sizeof(char_data_t)*NCHARS);
	if(in_mesetas) write_num(strtod(in_mesetas,0),savefile+MESETA_ADR,3);

	fp = fopen(outfile, "wb");
	assert(fp);
	fwrite(savefile, filesize, 1, fp);

	fclose(fp);
	
	return 0;
}
